/*
 * Mutex.c
 *
 * Copyright 2001, 2002 by Anthony Howe.  All rights reserved.
 *
 * Ralf S. Engelschall's MM Library, while nice, appears to be incomplete:
 * no POSIX semaphores, permission issues for SysV, and constant need to
 * lock/unlock sections before doing a memory allocation or free. The last
 * concerns me greatly, because some specialised routines really want
 * monitor-like behaviour, only one user in the entire routine at a time.
 */

/*
 * Choice of API to be used for serialization (semaphores & mutexes).
 */

#define SYSTEMV_API		1
#define POSIX_API		2
#define FCNTL_API		3
#define FLOCK_API		4
#define WIN32_API		5

#ifndef SERIALIZATION_API
#define SERIALIZATION_API	SYSTEMV_API
#endif

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "Mutex.h"

#if SERIALIZATION_API == FCNTL_API
#include <unistd.h>
struct mutex {
	int fd;
	struct flock lock;
	struct flock unlock;
	char lockfile[1];
};
#endif

#if SERIALIZATION_API == FLOCK_API
#include <sys/file.h>

#ifndef LOCK_SH
/* Operations for flock(). */
#define LOCK_SH		1    /* Shared lock.  */
#define LOCK_EX		2    /* Exclusive lock.  */
#define LOCK_UN		8    /* Unlock.  */
#endif

struct mutex {
	int fd;
	pid_t pid;
	char lockfile[1];
};
#endif

#if SERIALIZATION_API == POSIX_API
#include <semaphore.h>
#endif

#if SERIALIZATION_API == SYSTEMV_API
#include <sys/ipc.h>
#include <sys/sem.h>

#if (defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)) || defined(__bsdi__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
/* union semun is defined by including <sys/sem.h> */
#else
/* according to X/OPEN we have to define it ourselves */
union semun {
     int val;                  /* value for SETVAL */
     struct semid_ds *buf;     /* buffer for IPC_STAT, IPC_SET */
     unsigned short *array;    /* array for GETALL, SETALL */
			       /* Linux specific part: */
     struct seminfo *__buf;    /* buffer for IPC_INFO */
};
#endif

struct mutex {
	int id;
	struct sembuf lock;
	struct sembuf unlock;
};
#endif

#if SERIALIZATION_API == WIN32_API
#include <windows.h>
struct mutex {
	HANDLE handle;
};
#endif


void *
MutexCreate(const char *lockfile)
{
#if SERIALIZATION_API == FCNTL_API
	struct mutex *mex;

	if ((mex = malloc(sizeof *mex + strlen(lockfile))) != (struct mutex *) 0) {
		mex->lock.l_whence = SEEK_SET;		/* from current point */
		mex->lock.l_start = 0;			/* -"- */
		mex->lock.l_len = 0;			/* until end of file */
		mex->lock.l_type = F_WRLCK;		/* set exclusive/write lock */
		mex->lock.l_pid = 0;			/* pid not actually interesting */

		mex->unlock.l_whence = SEEK_SET;	/* from current point */
		mex->unlock.l_start = 0;		/* -"- */
		mex->unlock.l_len = 0;			/* until end of file */
		mex->unlock.l_type = F_UNLCK;		/* set exclusive/write lock */
		mex->unlock.l_pid = 0;			/* pid not actually interesting */

		strcpy(mex->lockfile, lockfile);

		(void) unlink(lockfile);

		if (0 <= (mex->fd = open(lockfile, O_CREAT|O_RDWR, 0600))) {
			return (void *) mex;
		}
		free(mex);
	}
#endif
#if SERIALIZATION_API == FLOCK_API
	struct mutex *mex;

	if ((mex = malloc(sizeof *mex + strlen(lockfile))) != (struct mutex *) 0) {
		strcpy(mex->lockfile, lockfile);
		(void) unlink(lockfile);

		if (0 <= (mex->fd = open(lockfile, O_CREAT|O_RDWR, 0600))) {
			mex->pid = getpid();
			return (void *) mex;
		}
		free(mex);
	}
#endif
#if SERIALIZATION_API == POSIX_API
	sem_t *sem;

	if ((sem = malloc(sizeof *sem)) != (sem_t *) 0) {
		if (sem_init(sem, 1, 1) == 0)
			return sem;
		free(sem);
	}
#endif
#if SERIALIZATION_API == SYSTEMV_API
	struct mutex *mex;
	union semun ick = { 0 };

	if ((mex = malloc(sizeof *mex)) != (struct mutex *) 0) {
		mex->lock.sem_op = 1;
		mex->lock.sem_num = 0;
		mex->lock.sem_flg = SEM_UNDO;

		mex->unlock.sem_op = -1;
		mex->unlock.sem_num = 0;
		mex->unlock.sem_flg = SEM_UNDO;

		mex->id = semget(IPC_PRIVATE, 1, IPC_CREAT|IPC_EXCL|S_IRUSR|S_IWUSR);
		if (mex->id < 0 && errno == EEXIST)
			mex->id = semget(IPC_PRIVATE, 1, IPC_EXCL|S_IRUSR|S_IWUSR);
		if (0 <= mex->id) {
			if (semctl(mex->id, 0, SETVAL, ick) == 0) {
				return (void *) mex;
			}
			/* ick is ignored here by IPC_RMID so says the man */
			(void) semctl(mex->id, 0, IPC_RMID, ick);
		}
		free(mex);
	}
#endif
#if SERIALIZATION_API == WIN32_API
{
	struct mutex *mex;

	if ((mex = malloc(sizeof *mex)) != (struct mutex *) 0) {
    		mex->handle = CreateMutex(NULL, FALSE, lockfile);

		if (mex->handle != (HANDLE) 0)
			return (void *) mex;

		free(mex);
	}
}
#endif
	return (void *) 0;
}


void
MutexDestroy(void *mp)
{
	if (mp == (void *) 0)
		return;

#if SERIALIZATION_API == FCNTL_API
	(void) close(((struct mutex *) mp)->fd);
	(void) unlink(((struct mutex *) mp)->lockfile);
	free(mp);
#endif
#if SERIALIZATION_API == FLOCK_API
	(void) close(((struct mutex *) mp)->fd);
	(void) unlink(((struct mutex *) mp)->lockfile);
	free(mp);
#endif
#if SERIALIZATION_API == POSIX_API
	errno = 0;
	do {
		(void) sem_destroy((sem_t *) mp);
	} while (errno == EBUSY);
	free(mp);
#endif
#if SERIALIZATION_API == SYSTEMV_API
{
	union semun ick = { 0 };
	(void) semctl(((struct mutex *) mp)->id, 0, IPC_RMID, ick);
	free(mp);
}
#endif
#if SERIALIZATION_API == WIN32_API
	CloseHandle(((struct mutex *) mp)->handle);
	free(mp);
#endif
}


int
MutexPermission(void *mp, int mode, int user, int group)
{
	if (mp == (void *) 0)
		return -1;

#if SERIALIZATION_API == FCNTL_API
	if (chmod(((struct mutex *) mp)->lockfile, mode) < 0)
		return -1;
	if (chown(((struct mutex *) mp)->lockfile, user, group) < 0)
		return -1;
#endif
#if SERIALIZATION_API == FLOCK_API
	if (chmod(((struct mutex *) mp)->lockfile, mode) < 0)
		return -1;
	if (chown(((struct mutex *) mp)->lockfile, user, group) < 0)
		return -1;
#endif
#if SERIALIZATION_API == POSIX_API
	/* Do nothing? */
#endif
#if SERIALIZATION_API == SYSTEMV_API
{
	union semun ick;
        struct semid_ds sembuf;

	ick.buf = &sembuf;
        if (semctl(((struct mutex *) mp)->id, 0, IPC_STAT, ick) != 0)
                return -1;
        sembuf.sem_perm.uid = user;
        sembuf.sem_perm.gid = group;
        sembuf.sem_perm.mode = mode;
        if (semctl(((struct mutex *) mp)->id, 0, IPC_SET, ick) != 0)
                return -1;
}
#endif
#if SERIALIZATION_API == WIN32_API
	/* Do nothing. */
#endif

	return 0;
}


int
MutexLock(void *mp)
{
	if (mp == (void *) 0)
		return -1;

#if SERIALIZATION_API == FCNTL_API
{
	struct mutex *mex = (struct mutex *) mp;

	for (errno = 0; fcntl(mex->fd, F_SETLKW, &mex->lock) != 0; ) {
		if (errno != EINTR)
			return -1;
	}
}
#endif
#if SERIALIZATION_API == FLOCK_API
{
	struct mutex *mex = (struct mutex *) mp;

	if (mex->pid != getpid()) {
		if ((mex->fd = open(mex->lockfile, O_RDWR, 0600)) < 0)
			return -1;
		mex->pid = getpid();
	}

	for (errno = 0; flock(mex->fd, LOCK_EX) != 0; ) {
		if (errno != EINTR)
			return -1;
	}
}
#endif
#if SERIALIZATION_API == POSIX_API
	for (errno = 0; sem_wait((sem_t *) mp) != 0; ) {
		if (errno != EINTR)
			return -1;
	}
#endif
#if SERIALIZATION_API == SYSTEMV_API
{
	struct mutex *mex = (struct mutex *) mp;

	for (errno = 0; semop(mex->id, &mex->lock, 1) != 0; ) {
		if (errno != EINTR)
			return -1;
	}
}
#endif
#if SERIALIZATION_API == WIN32_API
	switch (WaitForSingleObject(((struct mutex *) mp)->handle, INFINITE)) {
	case WAIT_TIMEOUT:
	case WAIT_OBJECT_0:
	case WAIT_ABANDONED:
		return 0;
	case WAIT_FAILED:
		return -1;
	}
#endif

	return 0;
}


int
MutexUnlock(void *mp)
{
	if (mp == (void *) 0)
		return -1;

#if SERIALIZATION_API == POSIX_API
	for (errno = 0; sem_post((sem_t *) mp) != 0; ) {
		if (errno != EINTR)
			return -1;
	}
#endif
#if SERIALIZATION_API == FCNTL_API
{
	struct mutex *mex = (struct mutex *) mp;

	for (errno = 0; fcntl(mex->fd, F_SETLKW, &mex->unlock) != 0; ) {
		if (errno != EINTR)
			return -1;
	}
}
#endif
#if SERIALIZATION_API == FLOCK_API
{
	struct mutex *mex = (struct mutex *) mp;

	if (mex->pid != getpid()) {
		if ((mex->fd = open(mex->lockfile, O_WRONLY, 0600)) < 0)
			return -1;
		mex->pid = getpid();
	}

	for (errno = 0; flock(mex->fd, LOCK_UN) != 0; ) {
		if (errno != EINTR)
			return -1;
	}
}
#endif
#if SERIALIZATION_API == SYSTEMV_API
{
	struct mutex *mex = (struct mutex *) mp;

	for (errno = 0; semop(mex->id, &mex->unlock, 1) != 0; ) {
		if (errno != EINTR)
			return -1;
	}
}
#endif
#if SERIALIZATION_API == WIN32_API
	if (ReleaseMutex(((struct mutex *) mp)->handle) == 0)
		return -1;
#endif

	return 0;
}

