/*
 * NetworkTable.c
 *
 * Copyright 2001, 2002 by Anthony Howe.  All rights reserved.
 */

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#include <stdlib.h>
#include <string.h>

#include "NetworkTable.h"

/*
 * Given a partial IP address, generate an IP address and mask.
 * This code is intended to handle IP6 and IP4 addresses.
 */
static const char *
setPartialIP(const char **address, struct in_addr *ip, struct in_addr *mask)
{
	int octet, length;
	const char *before, *here;
	unsigned char *ipp, *maskp;
	struct in_addr ipbuf, maskbuf;

	memset(&ipbuf, 0, sizeof ipbuf);
	memset(&maskbuf, 0, sizeof maskbuf);

	ipp = (unsigned char *) &ipbuf;
	maskp = (unsigned char *) &maskbuf;

	for (length = 0, here = *address; *here != '\0'; ++length) {
		octet = (int) strtol(before = here, (char **) &here, 10);

		if (before == here)
			return "Invalid IP address.";

		if (octet < 0 || 255 < octet)
			return "IP address octets must be between 0..255";

		if (*here == '.')
			++here;
		else if (*here == '/')
			break;
		else if (*here != '\0')
			return "Invalid IP address.";

		if (sizeof (struct in_addr) <= length)
			return "IP address too large.";

		*maskp++ = 0xff;
		*ipp++ = octet;
	}

	if (ip != (struct in_addr *) 0)
		*ip = ipbuf;

	if (mask != (struct in_addr *) 0)
		*mask = maskbuf;

	*address = here;

	return (char *) 0;
}

static int
ntCompare(const void *aa, const void *bb)
{
	int cmp;
	struct network *a = (struct network *) aa;
	struct network *b = (struct network *) bb;

	/* Sort netmasks in decending order. */
	if ((cmp = memcmp(&a->mask, &b->mask, sizeof a->mask)) != 0)
		return -cmp;

	/* Sort the IPs within a netmask in acending order. */
	return memcmp(&a->ip, &b->ip, sizeof a->ip);
}

static void
ntApplyMask(struct in_addr *ip, struct in_addr *mask)
{
	int i;

	for (i = 0; i < sizeof *mask; ++i)
		((unsigned char *) ip)[i] &= ((unsigned char *) mask)[i];
}

static void
ntFillMask(struct in_addr *mask, int bits)
{
	unsigned char *maskp = (unsigned char *) mask;
	unsigned char *stop = maskp + sizeof *mask;

	for ( ; maskp < stop; ++maskp) {
		if (bits < 8)
			break;
		*maskp = 0xff;
		bits -= 8;
	}

	*maskp++ = ~0 << (8 - bits);

	for ( ; maskp < stop; ++maskp)
		*maskp = 0;
}

int
ntAddIpMask(array_header *table, struct in_addr *ip, struct in_addr *mask)
{
	struct network *entry;

	entry = (struct network *) ap_push_array(table);
	if (entry == (struct network *) 0)
		return -1;

	entry->ip = *ip;
	entry->mask = *mask;

	qsort(table->elts, table->nelts, table->elt_size, ntCompare);

	return 0;
}

const char *
ntAddNetwork(array_header *table, const char *network)
{
	int bits;
	const char *error;
	struct in_addr ip, mask;

	if ((error = setPartialIP(&network, &ip, &mask)) != (const char *) 0)
		return error;

	if (*network == '/') {
		network++;
		bits = (int) strtol(network, (char **) &network, 10);

		if (*network == '\0') {
			/* a.b.c.d/nn */
			if (bits < 0 || sizeof mask * 8 < bits)
				return "Invalid network mask.";
			ntFillMask(&mask, bits);
		} else {
			/* a.b.c.d/A.B.C.D */
			error = setPartialIP(&network, &mask, (struct in_addr *) 0);
			if (error != (const char *) 0)
				return "Invalid network mask.";
		}

	}

	if (*network != '\0')
		return "Invalid network specifier.";

	if (ntAddIpMask(table, &ip, &mask))
		return "Failed to add IP/mask to network table.";

	return (char *) 0;
}

/*
 * Linear search of an array in decending order by netmask.
 */
struct network *
ntGetNetwork(array_header *table, struct in_addr *ip)
{
	int i;
	struct in_addr copy;
	struct network *array;

	if (table == (array_header *) 0)
		return (struct network *) 0;

	array = (struct network *) table->elts;

	for (i = 0; i < table->nelts; ++i) {
		copy = *ip;
		ntApplyMask(&copy, &array[i].mask);
		if (memcmp(&copy, &array[i].ip, sizeof copy) == 0)
			return &array[i];
	}

	return (struct network *) 0;
}

int
ntIsMember(array_header *table, struct in_addr *ip)
{
	return ntGetNetwork(table, ip) != (struct network *) 0;
}

