/*
 * SharedHash.c
 *
 * Shared memory hash table with long term disk storage.
 *
 * Copyright 2001, 2002 by Anthony Howe.  All rights reserved.
 */

/*
 * Maximum number of linear probes when searching the hash table
 * before swapping the oldest entry out to disk. I recommend a
 * small number betweeb 3 and 20. I think it is better to swap
 * out to disk an entry than spend too much time searching.
 *
 * This could be a module configuration option, but I am of the
 * opinion too many choices can result in poor performance. Best
 * to keep this as a small and reasonable constant.
 */
#ifndef MAX_PROBES
#define MAX_PROBES		8
#endif

/*
 * The average length of a key string. Used in computing extra
 * space required to store key strings in shared memory. When
 * key string space is exhausted, some hash entries are flushed
 * to disk to make room for newer entries.
 */
#ifndef AVERAGE_KEY_LENGTH
#define AVERAGE_KEY_LENGTH	20
#endif

/*
 * The longest supported key string. This should account for IP
 * addresses in ASCII, user names, virtual host names like
 * www.domain.tld and subdomain.domain.tld.  From what I recall,
 * domain names can be 22 characters plus a dot plus a 4 character
 * top-level-domain. I've allowed for extra space to account for
 * machine or sub-domain names.
 */
#ifndef MAX_KEY_LENGTH
#define MAX_KEY_LENGTH		128
#endif

#define ENABLE_LOG		1
#define BOUNDARY_CHECKING	1

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#include "httpd.h"
#include "http_log.h"
#include "http_conf_globals.h"
extern server_rec *watchMainServer;

#include "Mutex.h"
#include "Shared.h"
#include "Memory.h"
#include "SharedHash.h"

const char shLockFile[] = "SharedHash.lock";

const char shScanFormat[] = "%lu %lu %lu %lu %hd %lf %lu %lu";
const char shPrintFormat[] = "%lu %lu %lu %lu %hd %.3f %lu %lu";

#ifdef BOUNDARY_CHECKING
char *
shVerifyString(struct shTable *tp, char *str)
{
	if ((char *) tp->shared <= str && str < (char *) tp->eshared) {
		if (str + strlen(str) < (char *) tp->eshared)
			return str;
	}

	return (char *) 0;
}
#endif

int
shLock(struct shTable *tp)
{
	if (tp == (struct shTable *) 0)
		return -1;

	return MutexLock(tp->mutex);
}

int
shUnlock(struct shTable *tp)
{
	if (tp == (struct shTable *) 0)
		return -1;

	return MutexUnlock(tp->mutex);
}

/*
 * Store a shared memory hash table entry to disk.
 */
static void
shStore(struct shTable *tp, struct shEntry *entry)
{
	FILE *fp;
	char *sharedKey;

	if (entry == (struct shEntry *) 0)
		return;

#ifdef BOUNDARY_CHECKING
	if ((sharedKey = shVerifyString(tp, entry->key)) == (char *) 0)
		return;
#else
	if ((sharedKey = entry->key) == (char *) 0)
		return;
#endif

	/* ASSUME the workdir string has a trailing directory separator. */
	strcpy(tp->pathname, tp->workdir);
	strncat(tp->pathname, sharedKey, MAX_KEY_LENGTH);

	if ((fp = fopen(tp->pathname, "w")) != (FILE *) 0) {
		(void) fprintf(
			fp, shPrintFormat,
			entry->ifInOctets,
			entry->ifOutOctets,
			entry->ifRequests,
			entry->ifDocuments,
			entry->ifActive,
			entry->ifOutRate,
			entry->periodOctets,
			entry->periodMarker
		);

		(void) fprintf(fp, "\n");
		fclose(fp);

		(void) chown(tp->pathname, tp->uid, tp->gid);
	}
}

/*
 * Flush the entier shared memeory hash table to disk, releasing
 * the memory assigned for all the key strings.
 */
void
shFlushAll(struct shTable *tp)
{
	struct shEntry *here, *stop;

	stop = tp->array + tp->size;

	for (here = tp->array; here < stop; ++here) {
		shStore(tp, here);
		MemoryFree(tp->memory, here->key);
		here->key = (char *) 0;
	}
}

/*
 * Flush to disk a consective series of shared memory hash table
 * entries either side of the given index.
 */
static void
shFlush(struct shTable *tp, int index)
{
	int count = 0;
	struct shEntry *here, *stop;

	tp->info->flushes++;
	stop = tp->array + tp->size;

	/* Flush forward until next null. */
	here = tp->array + (index + 1) % tp->size;

	while (here->key != (char *) 0) {
		shStore(tp, here);
		MemoryFree(tp->memory, here->key);
		here->key = (char *) 0;
		count++;

		if (stop <= ++here)
			here = tp->array;
	}

	/* Flush backward until previous null. */
	here = tp->array + index;

	while (here->key != (char *) 0) {
		shStore(tp, here);
		MemoryFree(tp->memory, here->key);
		here->key = (char *) 0;
		count++;

		if (--here < tp->array)
			here = stop - 1;
	}

	if (count <= 0)
		shFlushAll(tp);
}

/*
 * Fetch a shared memory hash table entry from disk. If we run out of
 * shared memory space, then flush some entries to disk before fetching
 * the entry for the key.
 */
static struct shEntry *
shFetch(struct shTable *tp, const char *key, int index)
{
	FILE *fp;
	char *sharedKey;
	int rc, keylen = (int) strlen(key) + 1;
	struct shEntry *entry = &tp->array[index];

#ifdef BOUNDARY_CHECKING
	sharedKey = shVerifyString(tp, entry->key);
#else
	sharedKey = entry->key;
#endif

	sharedKey = MemoryResize(tp->memory, sharedKey, keylen);
	if (sharedKey == (char *) 0) {
		shFlush(tp, index);
		sharedKey = MemoryAllocate(tp->memory, keylen);
		if (sharedKey == (char *) 0) {
			return (struct shEntry *) 0;
		}
	}

	/* ASSUME the workdir string has a trailing directory separator. */
	strcpy(tp->pathname, tp->workdir);
	strncat(tp->pathname, key, MAX_KEY_LENGTH);
	strcpy(sharedKey, key);

	if ((fp = fopen(tp->pathname, "r")) != (FILE *) 0) {
		rc = fscanf(
			fp, shScanFormat,
			&entry->ifInOctets,
			&entry->ifOutOctets,
			&entry->ifRequests,
			&entry->ifDocuments,
			&entry->ifActive,
			&entry->ifOutRate,
			&entry->periodOctets,
			&entry->periodMarker
		);

		fclose(fp);
	}

	entry->key = sharedKey;

	return entry;
}

unsigned long
shHashCode(unsigned long hash, const char *k)
{
	if (k != (const char *) 0) {
		for ( ; *k != '\0'; ++k)
			hash = (hash * 37) + *k;
	}

	return hash;
}

int
shContainsKey(struct shTable *tp, const char *key)
{
	int i;
	struct shEntry *array;
	unsigned long hash, start;

	if (tp == (struct shTable *) 0 || key == (const char *) 0)
		return 0;

	array = tp->array;
	start = hash = shHashCode(0, key) % tp->size;

	for (i = 0; i <= MAX_PROBES; ++i) {
		if (array[hash].key == (char *) 0)
			return 0;
		if (strcmp(key, array[hash].key) == 0)
			return 1;

		/* Linear probe through the table. */
		hash = (hash + 1) % tp->size;
	}

	return 0;
}

/*
 * Return a pointer to the shared memory hash table entry for the key.
 */
struct shEntry *
shGetLockedEntry(struct shTable *tp, const char *key)
{
	int i;
	const char *k;
	struct shEntry *entry;
	unsigned long hash, start;

	if (tp == (struct shTable *) 0 || key == (const char *) 0)
		return (struct shEntry *) 0;

	/* A simple hash function. Factors of either 31 or 37 are good
	 * values for text strings. If the table size is prime, then
	 * we can double-hash or linear probe through entire table.
	 */
	start = hash = shHashCode(0, key) % tp->size;

	if (MutexLock(tp->mutex))
		return (struct shEntry *) 0;

	/* Perform a limited number of linear probes; no point in
	 * looking too far, better just to swap an entry out to disk.
	 */
	for (i = 0; i <= MAX_PROBES; ++i) {
		/* Free slot in the hash table? */
		if (tp->array[hash].key == (char *) 0) {
			if ((entry = shFetch(tp, key, hash)) == (struct shEntry *) 0) {
#ifdef ENABLE_LOG
				ap_log_error(
					APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, watchMainServer,
					"shFetch(%lx, key=\"%s\", hash=%d) failed filling empty entry",
					(unsigned long) tp, key, hash
				);
#endif
				MutexUnlock(tp->mutex);
			} else {
				tp->info->found++;
			}

			return entry;
		}

#ifdef BOUNDARY_CHECKING
		if (shVerifyString(tp, tp->array[hash].key) == (char *) 0) {
#ifdef ENABLE_LOG
			ap_log_error(
				APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, watchMainServer,
				"shVerifyString(%lx, %lx) failed!",
				(unsigned long) tp,
				(unsigned long) tp->array[hash].key
			);
#endif
			MutexUnlock(tp->mutex);
			return (struct shEntry *) 0;
		}
#endif

		/* Found existing key? */
		if (strcmp(key, tp->array[hash].key) == 0) {
			tp->info->found++;
			return &tp->array[hash];
		}

		/* Linear probe through the table. */
		hash = (hash + 1) % tp->size;
		tp->info->probes++;
	}

	/* No matching key found within probing limit. Replace the
	 * initial hash starting point.
	 */
	shStore(tp, &tp->array[start]);
	if ((entry = shFetch(tp, key, start)) == (struct shEntry *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, watchMainServer,
			"shFetch(%lx, key=\"%s\", start=%d) failed replacing entry",
			(unsigned long) tp, key, start
		);
#endif
		MutexUnlock(tp->mutex);
	} else {
		tp->info->faults++;
	}

	return entry;
}

/*
 * Create a shared memory table.
 */
struct shTable *
shCreate(struct pool *p, int size, const char *workdir, uid_t uid, gid_t gid)
{
	size_t bytes;
	const char *lockfile;
	struct shTable *tp;

	if (size <= MAX_PROBES) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, watchMainServer,
			"shCreate() hash table size too small (%d)",
			size
		);
#endif
		goto error0;
	}

	if (workdir == (const char *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, watchMainServer,
			"shCreate() workdir argument cannot be NULL",
			size
		);
#endif
		goto error0;
	}

	lockfile = ap_pstrcat(p, workdir, shLockFile, (char *) 0);
	if (lockfile == (const char *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"shCreate() failed to allocate lockfile string"
		);
#endif
		goto error0;
	}

	/* Allocate enough storage for the table management elements,
	 * which includes a pathname construction buffer.
	 */
	tp = ap_pcalloc(p, sizeof *tp + strlen(workdir) + 1 + MAX_KEY_LENGTH + 1);
	if (tp == (struct shTable *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"shCreate() failed to allocate shTable structure"
		);
#endif
		goto error0;
	}

	/* Space for global data. */
	bytes = sizeof *tp->info;

	/* We need enough room for an array of N entries, with an
	 * average key length of M bytes.
	 */
	bytes += size * (sizeof *tp->array + AVERAGE_KEY_LENGTH);

	if ((tp->shared = SharedCreate(bytes)) == (void *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"SharedCreate(%lu) failed", bytes
		);
#endif
		goto error0;
	}

	if (SharedPermission(tp->shared, 0600, ap_user_id, ap_group_id) < 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"SharedPermission(%lx, 0600, %d=%s, %d=%s) failed",
			(unsigned long) tp->shared,
			ap_user_id, watchGetUidName(p, ap_user_id),
			ap_group_id, watchGetGidName(p, ap_group_id)
		);
#endif
		goto error1;
	}

	if ((tp->memory = MemoryCreate(tp->shared, bytes)) == (void *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"MemoryCreate(%lx, %lu) failed",
			(unsigned long) tp->shared, bytes
		);
#endif
		goto error1;
	}

	if ((tp->mutex = MutexCreate(lockfile)) == (void *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"MutexCreate(%s) failed", lockfile
		);
#endif
		goto error1;
	}

	if (MutexPermission(tp->mutex, 0600, ap_user_id, ap_group_id) < 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"MutexPermission(%lx, 0600, %d=%s, %d=%s) failed",
			(unsigned long) tp->mutex,
			ap_user_id, watchGetUidName(p, ap_user_id),
			ap_group_id, watchGetGidName(p, ap_group_id)
		);
#endif
		goto error2;
	}

	tp->info = (struct shInfo *) MemoryAllocate(tp->memory, sizeof *tp->info);
	if (tp->info == (struct shInfo *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"MemoryAllocate(%lx, %lu) #1 failed",
			(unsigned long) tp->memory, sizeof *tp->info
		);
#endif
		goto error3;
	}

	tp->array = (struct shEntry *) MemoryAllocate(tp->memory, size * sizeof *tp->array);
	if (tp->array == (struct shEntry *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"MemoryAllocate(%lx, %lu) #2 failed",
			(unsigned long) tp->memory, size * sizeof *tp->array
		);
#endif
		goto error3;
	}

	MemorySet(tp->info, 0);
	MemorySet(tp->array, 0);

	tp->pathname = (char *) tp + sizeof *tp;
	tp->eshared = tp->shared + bytes;
	tp->workdir = workdir;
	tp->size = size;
	tp->uid = uid;
	tp->gid = gid;

	return tp;
error3:
	MutexDestroy(tp->mutex);
error2:
	MemoryDestroy(tp->memory);
error1:
	SharedDestroy(tp->shared);
error0:
	return (struct shTable *) 0;
}

/*
 * Save the entrie shared memory array to disk.
 */
void
shDestroy(void *data)
{
	struct shTable *tp = (struct shTable *) data;

	if (tp != (struct shTable *) 0) {
		shFlushAll(tp);
		MutexDestroy(tp->mutex);
		MemoryDestroy(tp->memory);
		SharedDestroy(tp->shared);
	}
}

