/*
 * mod_watch.c for Apache 1.3
 *
 * Copyright 2001, 2002 by Anthony Howe.  All rights reserved.
 *
 *
 * LICENSE
 * -------
 *
 * 1) Non-commercial redistribution and use of the original source
 * distribution, without modification, is permitted provided you retain
 * this notice, the author's copyright notice above, and the disclaimer
 * below.
 *
 * 2) Commercial use, in the role of "end-user" for internal business
 * operations, of the original source distribution, without modification,
 * is permitted provided you retain this notice, the author's copyright
 * notice above, and the disclaimer below.
 *
 * 3) Commercial redistribution (including but not limited to bundling,
 * collections, resale) of the original source distribution and/or derived
 * binaries, with or without modification, is permitted with a) the prior
 * written permission from the author, b) financial remuneration to
 * the author, c) the above copyright notice is retained and reproduced by
 * the binary, and d) the disclaimer below appears in the documentation
 * and/or other materials provided with the distribution.
 *
 *
 * DISCLAIMER
 * ----------
 *
 * THIS SOFTWARE IS PROVIDE "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO WAY SHALL THE
 * AUTHOR OR LICENSEE BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef SPOOLDIR
#define SPOOLDIR		"/var/spool/mod_watch/"
#endif

/*
 * The size of the hash table should be a small prime number:
 *
 *	449, 673, 991, 1409, 2411
 *
 * Using a prime number for the table size means that double
 * hashing or linear-probing can visit all possible entries.
 *
 * This is NOT a runtime option, because its not something I
 * want people to play with unless absolutely necessary.
 */
#ifndef TABLE_SIZE
#define TABLE_SIZE		991
#endif

/*
 * The minimum amount of time in seconds between average rate
 * calculations.
 */
#ifndef SECONDS_PER_PERIOD
#define SECONDS_PER_PERIOD	300
#endif

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#define MODULE			"mod_watch"
#define AUTHOR			"achowe@snert.com"
#define VERSION			"3.17"
#define COPYRIGHT_DATE		"2001, 2002"

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_conf_globals.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "ap_config.h"

#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>

module MODULE_VAR_EXPORT watch_module;

/* This is NOT static for a reason so that it can
 * be used for error reporting else where.
 */
server_rec *watchMainServer;

/***********************************************************************
 *** Macros
 ***********************************************************************/

#define UNSET			-1
#define MERGE(p, c)		(c == UNSET ? p : c)
#define MERGE_PTR(p, c)		((void *) c == (void *) 0 ? p : c)

/***********************************************************************
 *** Constants
 ***********************************************************************/

#define WATCH_SERVER			1
#define WATCH_SERVER_BY_PORT		2

#define WATCH_VHOST			4
#define WATCH_VHOST_BY_PORT		8

#define WATCH_OWNER			16
#define WATCH_OWNER_BY_PORT		32

#define WATCH_REMOTE_IP			64
#define WATCH_REMOTE_IP_BY_PORT		128

#define WATCH_HEADERS_IN		256
#define WATCH_HEADERS_OUT		512
#define WATCH_CONTENT_LENGTH		1024

#define WATCH_DEFAULT			32768
#define WATCH_ALL			(~0)

#define WATCH_INFO		1
#define WATCH_LIST		2
#define WATCH_TABLE		4
#define WATCH_COPYRIGHT		8
#define WATCH_FLUSH		16

/***********************************************************************
 *** String Constants
 ***********************************************************************/

static const char SERVER[] = "SERVER";
static const char LOCKFILE[] = "mod_watch.lock";
static const char ISHANDLER[] = "watch-is-handler";
static const char STARTTIME[] = "watch-start-time";

static const char watch_all[] = "all";
static const char watch_default[] = "default";
static const char watch_flush[] = "watch-flush";
static const char watch_info[] = "watch-info";
static const char watch_list[] = "watch-list";
static const char watch_table[] = "watch-table";
static const char watch_copyright[] = "watch-copyright";

static const char text_plain[] = "text/plain";
static const char text_html[] = "text/html";

/***********************************************************************
 *** Global variables - should be read-only by child processes.
 ***********************************************************************/

typedef struct watchConfDir {
	int index;
} watchConfDir;

static int watch_what = WATCH_DEFAULT | WATCH_VHOST | WATCH_SERVER;
static int watch_not_logged = 0;
static int watchModVhostAlias = 0;

static char *spoolDirectoryRun;
static char *spoolDirectoryConf;

static array_header *watchSpoolFileConf;
static array_header *watchSpoolFileList;

static array_header *watchDocumentsConf;
static array_header *watchDocumentsList;
static const char *watchDocumentsDef[] = {
	text_html,
	text_plain,
	"text/xml",
	"application/pdf",
	"text/enriched",
	"text/richtext",
	"application/*word",
	"application/*excel",
	"application/postscript",
	"application/rtf",
/*	"httpd/unix-directory", */
	(char *) 0
};

/***********************************************************************
 *** Shared Memory Hash Table Routines
 ***********************************************************************/

#include "SharedHash.h"

static struct shTable *shtable;
static int shTableSize = TABLE_SIZE;

/***********************************************************************
 *** Semaphore / Mutex Routines
 ***********************************************************************/

#include "Mutex.h"

static void *mutex;

/***********************************************************************
 *** Network Table Routines
 ***********************************************************************/

#include "NetworkTable.h"

static array_header *networkIncludeConf;
static array_header *networkIncludeList;
static array_header *networkExcludeConf;
static array_header *networkExcludeList;

/***********************************************************************
 *** Storage Policy Prototypes & Types
 ***********************************************************************/

typedef struct storage_policy {
	const char *name;
	int (*content) (request_rec *);
	void (*post) (request_rec *, const char *, const char *);
	void (*log) (request_rec *, const char *, struct shEntry *);
} storage_policy;

static int watchContentInfoFiles(request_rec *r);
static int watchContentInfoHash(request_rec *r);

static void watchPostFiles(request_rec *r, const char *tkey, const char *hkey);
static void watchPostHash(request_rec *r, const char *tkey, const char *hkey);

static void watchLogFiles(request_rec *r, const char *key, struct shEntry *);
static void watchLogHash(request_rec *r, const char *key, struct shEntry *);

#define STORAGE_HASH		0
#define STORAGE_FILES		1

static storage_policy policy_table[] = {
	{ "hash", watchContentInfoHash, watchPostHash, watchLogHash },
	{ "files", watchContentInfoFiles, watchPostFiles, watchLogFiles },
	{ 0, 0 }
};

static storage_policy *watchStoragePolicy;

/***********************************************************************
 *** Support Routines
 ***********************************************************************/

#define UNKNOWN_UID		((uid_t) ~0)
#define UNKNOWN_GID		((gid_t) ~0)

/*
 */
const char *
watchGetUidName(struct pool *p, uid_t uid)
{
	if (uid == UNKNOWN_UID)
		return "unknown-user";

#if defined(WIN32) || defined(NETWARE)
	return "default-user";
#else
{
	struct passwd *pw;

	if ((pw = getpwuid(uid)) != (struct passwd *) 0)
		return ap_pstrdup(p, pw->pw_name);

	return "unknown-user";
}
#endif
}

const char *
watchGetGidName(struct pool *p, gid_t gid)
{
	if (gid == UNKNOWN_GID)
		return "unknown-group";

#if defined(WIN32) || defined(NETWARE)
	return "default-group";
#else
{
	struct group *gr;

	if ((gr = getgrgid(gid)) != (struct group *) 0)
		return ap_pstrdup(p, gr->gr_name);

	return "unknown-group";
}
#endif
}

int
mkpath(struct pool *p, char *dir)
{
	int i, count;
	char *prefix;
	struct stat sb;

	if (!ap_os_is_path_absolute(dir)) {
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"%s is not an absolute path", dir
		);
		return -1;
	}

	if ((prefix = ap_palloc(p, strlen(dir) + 1)) == (char *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"mkpath(): ap_palloc() failed", dir
		);
		return -1;
	}

	count = ap_count_dirs(dir) + 1;
	for (i = 2; i <= count; ++i) {
		(void) ap_make_dirstr_prefix(prefix, dir, i);
		if (stat(prefix, &sb) < 0) {
			if (mkdir(prefix, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH)) {
				ap_log_error(
					APLOG_MARK, APLOG_ERR, watchMainServer,
					"failed to create directory %s", prefix
				);
				return -1;
			}
			if (chown(prefix, ap_user_id, ap_group_id)) {
				ap_log_error(
					APLOG_MARK, APLOG_ERR, watchMainServer,
					"failed to chown %s to %d:%d",
					prefix, ap_user_id, ap_group_id
				);
				return -1;
			}
		} else if (!S_ISDIR(sb.st_mode)) {
			ap_log_error(
				APLOG_MARK, APLOG_ERR, watchMainServer,
				"%s must be a directory", prefix
			);
			return -1;
		}
	}

	/* The last directory segment must be owned by the Apache user. */
	if (sb.st_uid != ap_user_id) {
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"%s must be owned by user ID %d", prefix, ap_user_id
		);
		return -1;
	}

	return 0;
}

static const char *
setDirectory(struct pool *p, char *arg, char **directory)
{
	struct stat buf;

        arg = ap_server_root_relative(p, arg);

	/* Append trailing slash. */
	if (strrchr(arg, '/')[1] != '\0')
		*directory = ap_pstrcat(p, arg, "/", (char *) 0);
	else
		*directory = ap_pstrdup(p, arg);

	return (const char *) 0;
}

static const char *
setUnsignedInt(const char *arg, unsigned int *number)
{
	char *stop;
	long value;

	if (arg == (char *) 0 || *arg == '\0')
		return "Integer unspecified";

	value = strtol(arg, &stop, 10);

	if (*stop != '\0')
		return "Not a decimal integer";

	if (value < 0)
		return "Not a positive integer";

	if (UINT_MAX < value)
		return "Integer too large";

	*number = (unsigned int) value;

	return (const char *) 0;
}

/*
 * Return the canonical name of the virtual host.
 *
 * The ap_get_server_name() returns either the defined ServerName
 * or the client supplied host name depending on the setting of
 * UseCanonicalName. When UseCanonicalName is on, only the virtual
 * host's ServerName is ever returned. When UseCanonicalName is off
 * the host name from a proxy URI or Host: header will be returned,
 * in which case there may be several different names per server.
 *
 * We always want to use the canonical name of a server when available,
 * regardless of the setting of UseCanonicalName or the presence of
 * mod_vhost_alias, so that we only track the "one true name".
 */
static char *
watchGetCanonicalName(request_rec *r)
{
	int i;
	char *host;
	char **list;
	array_header *names;

	if (r->hostname == (char *) 0)
		return (char *) 0;

	/* Does request host name match the ServerName? */
	if (strcasecmp(r->hostname, r->server->server_hostname) == 0)
		return r->server->server_hostname;

	/* Does request host name match a ServerAlias for a <VirtualHost>. */
	names = r->server->names;
	if (names != (array_header *) 0) {
		list = (char **) names->elts;
		for (i = 0; i < names->nelts; ++i) {
			if (list[i] != (char *) 0 && strcasecmp(r->hostname, list[i]) == 0)
				return r->server->server_hostname;
		}
	}

	names = r->server->wild_names;
	if (names != (array_header *) 0) {
		list = (char **) names->elts;
		for (i = 0; i < names->nelts; ++i) {
			if (list[i] != (char *) 0 && ap_strcasecmp_match(r->hostname, list[i]) == 0)
				return r->server->server_hostname;
		}
	}

	if (watchModVhostAlias) {
		/* The client supplied host name does not correspond to a
		 * declared name or alias for this server. See if a simple
		 * sub-request for the host name in question would succeed,
		 * in which case the client supplied host name is good.
		 */
		int exists;
		request_rec *sub;

		sub = ap_sub_req_method_uri("HEAD", "/", r);
		exists = (sub->finfo.st_mode != 0);
		ap_destroy_sub_req(sub);

		ap_log_error(
			APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, watchMainServer,
			"watchGetCanonicalName(%lx) host=\"%s\" exits=%d",
			r, r->hostname, exists
		);

		if (exists) {
			ap_str_tolower((char *) r->hostname);
			return (char *) r->hostname;
		}
	}

	/* The client supplied host name does not correspond to ANYTHING.
	 * Do NOT log the results of this request by virtual host, since
	 * we cannot determine the host name.
	 */

	return (char *) 0;
}

/***********************************************************************
 ***  Create, Init, and Merge Routines
 ***********************************************************************/

static void *
watchCreateDir(struct pool *p, char *dir)
{
	struct watchConfDir *conf = ap_pcalloc(p, sizeof *conf);

	conf->index = UNSET;

	return (void *) conf;
}

/*
 *
 */
static void
watchInit(server_rec *s, struct pool *p)
{
	/* Handy to have for error logging. */
	watchMainServer = s;

	/* Is mod_vhost_alias in play? */
	watchModVhostAlias = (ap_find_linked_module("mod_vhost_alias.c") != (module *) 0);

        ap_add_version_component(MODULE"/"VERSION);

	/* If no WatchDocuments specified, use defaults. */
	if (watchDocumentsConf == (array_header *) 0) {
		const char **list, **mime;

		watchDocumentsConf = ap_make_array(
			p,
			sizeof watchDocumentsDef / sizeof *watchDocumentsDef,
			sizeof *watchDocumentsDef
		);

		for (list = watchDocumentsDef; *list != (char *) 0; ++list) {
			mime = (const char **) ap_push_array(watchDocumentsConf);
			if (mime == (const char **) 0)
				exit(APEXIT_INIT);
			*mime = *list;
		}
	}

	if (spoolDirectoryConf == (char *) 0) {
		const char *msg = setDirectory(p, SPOOLDIR, &spoolDirectoryConf);
		if (msg != (const char *) 0) {
			ap_log_error(
				APLOG_MARK, APLOG_CRIT, watchMainServer,
				"%s: %s", msg, SPOOLDIR
			);
			exit(APEXIT_INIT);
		}
	}

	if (mkpath(p, spoolDirectoryConf))
		exit(APEXIT_INIT);

	/* Prepare for next restart. */
        watchDocumentsList = watchDocumentsConf;
        watchDocumentsConf = (array_header *) 0;

	networkIncludeList = networkIncludeConf;
	networkIncludeConf = (array_header *) 0;

	networkExcludeList = networkExcludeConf;
	networkExcludeConf = (array_header *) 0;

	watchSpoolFileList = watchSpoolFileConf;
	watchSpoolFileConf = (array_header *) 0;

	spoolDirectoryRun = spoolDirectoryConf;
	spoolDirectoryConf = (char *) 0;

	if (watchStoragePolicy == &policy_table[STORAGE_FILES]) {
		/* Create a mutex used to control access to the weenie files. */
		const char *lockfile = ap_pstrcat(p, spoolDirectoryRun, LOCKFILE, (char *) 0);
		if (lockfile == (const char *) 0) {
			ap_log_error(
				APLOG_MARK, APLOG_ERR, watchMainServer,
				"watchInit() failed to allocate lockfile string"
			);
			exit(APEXIT_INIT);
		}

		if ((mutex = MutexCreate(lockfile)) == (void *) 0) {
			ap_log_error(
				APLOG_MARK, APLOG_ERR, watchMainServer,
				"MutexCreate(%s) failed in watchInit()", lockfile
			);
			exit(APEXIT_INIT);
		}

		if (MutexPermission(mutex, 0600, ap_user_id, ap_group_id) < 0) {
			ap_log_error(
				APLOG_MARK, APLOG_ERR, watchMainServer,
				"MutexPermission(%lx, 0600, %d=%s, %d=%s) failed in watchInit()",
				(unsigned long) mutex,
				ap_user_id, watchGetUidName(p, ap_user_id),
				ap_group_id, watchGetGidName(p, ap_group_id)
			);
			exit(APEXIT_INIT);
		}
		ap_register_cleanup(p, mutex, MutexDestroy, ap_null_cleanup);
	} else {
		watchStoragePolicy = &policy_table[STORAGE_HASH];

		/* Create shared memory hash table. */
		shtable = shCreate(p, shTableSize, spoolDirectoryRun, ap_user_id, ap_group_id);
		if (shtable == (struct shTable *) 0) {
			ap_log_error(
				APLOG_MARK, APLOG_CRIT, watchMainServer,
				"shCreate(%lx, %lu, %s) failed in watchInit()",
				(unsigned long) p, shTableSize, spoolDirectoryRun
			);
			exit(APEXIT_INIT);
		}
		ap_register_cleanup(p, shtable, shDestroy, ap_null_cleanup);
	}

	ap_log_error(
		APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, watchMainServer,
		"watchInit(%lx, %lx) watchStoragePolicy=%lx",
		(unsigned long) s, (unsigned long) p, (unsigned long) watchStoragePolicy
	);
}

/***********************************************************************
 ***  Handlers
 ***********************************************************************/

static int
watchHeaderLength(void *data, const char *key, const char *value)
{
	unsigned long *octets = (unsigned long *) data;

	/* Count length of the header plus account for a colon,
	 * space, and CRLF.
	 */
	*octets += strlen(key) + strlen(value) + 4;

	return 1;
}

static struct shEntry *
watchCounters(request_rec *r)
{
	struct shEntry *data;
	const char *value, *gzip;
	unsigned long content_length, hdr_out;

	data = ap_palloc(r->pool, sizeof *data);

	data->key = (char *) 0;

	data->ifInOctets = 0;
	if (watch_what & WATCH_HEADERS_IN) {
		/* Count the request plus CRLF. */
		data->ifInOctets = strlen(r->the_request) + 2;
		ap_table_do(watchHeaderLength, &data->ifInOctets, r->headers_in, (char *) 0);
	}

	if ((watch_what & WATCH_CONTENT_LENGTH) && r->method_number == M_POST) {
		/* Assume content-length input bytes were consumed
		 * by a module or CGI. There is no way to guarantee
		 * that these bytes were actually sent and read -
		 * someone could easily falsify the headers by hand
		 * just to spike the graph.
		 */
		value = ap_table_get(r->headers_in, "Content-Length");
		if (value != (const char *) 0)
			data->ifInOctets += strtol(value, (char **) 0, 10);
	}

	hdr_out = 0;
	if (watch_what & WATCH_HEADERS_OUT)
		ap_table_do(watchHeaderLength, &hdr_out, r->headers_out, (char *) 0);

	/* Size of response sent. */
	data->ifOutOctets = hdr_out + r->bytes_sent;

	/* Check for mod_gzip and use the compressed output size. */
 	gzip = ap_table_get(r->notes, "mod_gzip_output_size");
 	if (gzip != (const char *) 0) {
 		unsigned long size = (unsigned long) strtol(gzip, (char **) 0, 10);
		data->ifOutOctets = hdr_out + (0 < size ? size : r->bytes_sent);
	}

	ap_log_error(
		APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, watchMainServer,
		"watchCounters(r=%lx) r->bytes_sent=%lu, mod_gzip_output_size=%s",
		(unsigned long) r, r->bytes_sent, gzip == (const char *) 0 ? "undefined" : gzip
	);

	data->ifRequests = 1;
	data->ifDocuments = 0;
	data->ifOutRate = 0;	/* not used, filler */
	data->ifActive = 1;	/* not used, filler */

	/* Update ifDocuments counter. */
	if (r->content_type != (const char *) 0) {
		int i;
		request_rec *rr;
		char **list = (char **) watchDocumentsList->elts;

		/* Find actual content handler issued. */
		for (rr = r; rr->next != (request_rec *) 0; rr = rr->next)
			;

		for (i = 0; i < watchDocumentsList->nelts; ++i) {
			if (ap_strcasecmp_match(rr->content_type, list[i]) == 0) {
				data->ifDocuments = 1;
				break;
			}
		}
	}

	return data;
}

/*
 * Pick a target, which is either the server name of the request
 * or a specified ~user name or ~SERVER.
 */
static const char *
watchTarget(request_rec *r)
{
	if (r->uri[0] == '/' && r->uri[1] == '~' && ap_isalnum(r->uri[2])) {
		const char *remainder = r->uri + 2;
		return ap_getword(r->pool, &remainder, '/');
	}

	return watchGetCanonicalName(r);
}

/*
 * This handler is used in Phase 0 to store the current values of the
 * counters BEFORE this request is completed in the request_rec->notes
 * table. This allows other modules to make use of these counters.
 *
 * Other 3rd party modules can obtain the values of the counters at the
 * start of Phase 1. The values are stored in the request_rec->notes
 * table with the following key names:
 *
 *	watch-file-owner
 *	watch-virtual-host
 *	watch-server
 *	watch-remote-ip
 *	watch-location
 *
 * The value returned from ap_table_get() must be cast to a structure
 * pointer. The structure should be declared as:
 *
 *	struct watchEntry {
 *		char *key;
 *		unsigned long ifInOctets;
 *		unsigned long ifOutOctets;
 *		unsigned long ifRequests;
 *		unsigned long ifDocuments;
 *		unsigned lomg ifOutRate;
 *		short         ifActive;
 *	};
 *
 * For example:
 *
 *	struct watchEntry *entry;
 *	entry = (struct watchEntry *) ap_table_get(r->notes, "watch-virtual-host");
 *
 * These values are a per request copy of the data stored by mod_watch,
 * and should be considered read-only so that other modules can benefit
 * from them. Changing the values here will not affect the originals
 * maintained by mod_watch.
 */
static int
watchPost(request_rec *r)
{
	char *key;
	struct watchConfDir *dconf;

	/* Do not count if the connection is excluded. */
	if (ntIsMember(networkExcludeList, &r->connection->remote_addr.sin_addr)
	&& !ntIsMember(networkIncludeList, &r->connection->remote_addr.sin_addr))
		return DECLINED;

	key = (char *) watchGetUidName(r->pool, r->finfo.st_uid);
	if (watch_what & WATCH_OWNER)
		(*watchStoragePolicy->post)(r, "watch-file-owner", key);

	if (watch_what & WATCH_OWNER_BY_PORT) {
		key = ap_psprintf(r->pool, "%s:%u", key, ap_get_server_port(r));
		(*watchStoragePolicy->post)(r, "watch-file-owner-by-port", key);
	}

	if ((key = watchGetCanonicalName(r)) != (char *) 0) {
		if (watch_what & WATCH_VHOST)
			(*watchStoragePolicy->post)(r, "watch-virtual-host", key);

		if (watch_what & WATCH_VHOST_BY_PORT) {
			key = ap_psprintf(r->pool, "%s:%u", key, ap_get_server_port(r));
			(*watchStoragePolicy->post)(r, "watch-virtual-host-by-port", key);
		}
	}

	if (watch_what & WATCH_SERVER)
		(*watchStoragePolicy->post)(r, "watch-server", SERVER);

	if (watch_what & WATCH_OWNER_BY_PORT) {
		key = ap_psprintf(r->pool, "%s:%u", SERVER, ap_get_server_port(r));
		(*watchStoragePolicy->post)(r, "watch-server-by-port", key);
	}

	if (watch_what & WATCH_REMOTE_IP)
		(*watchStoragePolicy->post)(r, "watch-remote-ip", r->connection->remote_ip);

	if (watch_what & WATCH_REMOTE_IP_BY_PORT) {
		key = ap_psprintf(r->pool, "%s:%u", r->connection->remote_ip, ap_get_server_port(r));
		(*watchStoragePolicy->post)(r, "watch-remote-ip-by-port", key);
	}

	dconf = (struct watchConfDir *) ap_get_module_config(r->per_dir_config, &watch_module);

	if (dconf != (struct watchConfDir *) 0 && dconf->index != UNSET)
		(*watchStoragePolicy->post)(r, "watch-location", ((char **) watchSpoolFileList->elts)[dconf->index]);

	return DECLINED;
}

/*
 * All of our content handlers return plain text (well almost).
 */
static int
watchFixup(request_rec *r)
{
	int *isWatchHandler = ap_palloc(r->pool, sizeof (int));

	ap_table_setn(r->notes, ISHANDLER, (char *) isWatchHandler);

	*isWatchHandler = 0;

	ap_log_error(
		APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, watchMainServer,
		"watchFixup(%lx) handler=%s", r, r->handler
	);

	if (r->handler == (char *) 0)
		return DECLINED;

	if (ap_strcasecmp_match(r->handler, watch_info) == 0) {
		*isWatchHandler = WATCH_INFO;
		r->content_type = text_plain;
	} else if (ap_strcasecmp_match(r->handler, watch_flush) == 0) {
		*isWatchHandler = WATCH_FLUSH;
		r->content_type = text_plain;
	} else if (ap_strcasecmp_match(r->handler, watch_list) == 0) {
		*isWatchHandler = WATCH_LIST;
		r->content_type = text_plain;
	} else if (ap_strcasecmp_match(r->handler, watch_table) == 0) {
		*isWatchHandler = WATCH_TABLE;
		r->content_type = text_html;
	} else if (ap_strcasecmp_match(r->handler, watch_copyright) == 0) {
		*isWatchHandler = WATCH_COPYRIGHT;
		r->content_type = text_plain;
	}

	return *isWatchHandler == 0 ? DECLINED : OK;
}

static int
watchLog(request_rec *r)
{
	char *key;
	int *isWatchHandler;
	struct shEntry *data;
	struct watchConfDir *dconf = (struct watchConfDir *) ap_get_module_config(
		r->per_dir_config, &watch_module
	);

	isWatchHandler = (int *) ap_table_get(r->notes, ISHANDLER);

	/* Do not count the watch-info reports. */
	if (isWatchHandler != (int *) 0 && *isWatchHandler != 0) {
		/* The Writing Apache Modules book on p78 states that all the
		 * log phase handlers are executed regardless of whether they
		 * return OK or DECLINED. A log handler can short-circuit the
		 * remaining log handlers if it returns something other than
		 * OK or DECLINED. This is highly dependant on module load
		 * order and could affect the correct opertaion of other
		 * modules if used.
		 *
		 * The cleaner way to do this is through the httpd.conf file:
		 *
		 *	# Disable the previous CustomLog directive.
		 * 	#CustomLog logs/access_log common
		 *	SetEnvIf Request_URI "/watch-info$" IGNORE
		 *	CustomLog /dev/null common env=IGNORE
		 *	CustomLog logs/access_log common env=!IGNORE
		 */
		return (*isWatchHandler & watch_not_logged) == 0 ? OK : r->status;
	}

	/* Do not count if the connection is excluded. */
	if (ntIsMember(networkExcludeList, &r->connection->remote_addr.sin_addr)
	&& !ntIsMember(networkIncludeList, &r->connection->remote_addr.sin_addr))
		return DECLINED;

	/* Find last sub-request for bytes actually sent. */
	for ( ; r->next != (request_rec *) 0; r = r->next)
		;

	data = watchCounters(r);

	key = (char *) watchGetUidName(r->pool, r->finfo.st_uid);
	if (watch_what & WATCH_OWNER)
		(*watchStoragePolicy->log)(r, key, data);
	if (watch_what & WATCH_OWNER_BY_PORT) {
		key = ap_psprintf(r->pool, "%s:%u", key, ap_get_server_port(r));
		(*watchStoragePolicy->log)(r, key, data);
	}

	if ((key = watchGetCanonicalName(r)) != (char *) 0) {
		if (watch_what & WATCH_VHOST)
			(*watchStoragePolicy->log)(r, key, data);
		if (watch_what & WATCH_VHOST_BY_PORT) {
			key = ap_psprintf(r->pool, "%s:%u", key, ap_get_server_port(r));
			(*watchStoragePolicy->log)(r, key, data);
		}
	}

	if (watch_what & WATCH_SERVER)
		(*watchStoragePolicy->log)(r, SERVER, data);
	if (watch_what & WATCH_SERVER_BY_PORT) {
		key = ap_psprintf(r->pool, "%s:%u", SERVER, ap_get_server_port(r));
		(*watchStoragePolicy->log)(r, key, data);
	}

	if (watch_what & WATCH_REMOTE_IP)
		(*watchStoragePolicy->log)(r, r->connection->remote_ip, data);
	if (watch_what & WATCH_REMOTE_IP_BY_PORT) {
		key = ap_psprintf(r->pool, "%s:%u", r->connection->remote_ip, ap_get_server_port(r));
		(*watchStoragePolicy->log)(r, key, data);
	}

	if (dconf != (struct watchConfDir *) 0 && dconf->index != UNSET)
		(*watchStoragePolicy->log)(r, ((char **) watchSpoolFileList->elts)[dconf->index], data);

	return DECLINED;
}

static void
watchContentInfoLine(request_rec *r, struct shEntry *entry)
{
	ap_rprintf(
		r, "%s %lu ",
		entry->key, r->request_time - ap_restart_time
	);

	ap_rprintf(
		r, shPrintFormat,
		entry->ifInOctets,
		entry->ifOutOctets,
		entry->ifRequests,
		entry->ifDocuments,
		entry->ifActive,
		entry->ifOutRate
	);
	ap_rprintf(r, "\n");
}

/*
 * Write the content, which is single text line containing the
 * target name, server uptime in seconds, octets in, octets out,
 * requests in, and documents out.
 */
static int
watchContentInfo(request_rec *r, struct shEntry *entry)
{
	ap_send_http_header(r);

	if (!r->header_only)
		watchContentInfoLine(r, entry);

	return OK;
}

static int
watchContentCopyright(request_rec *r)
{
	ap_send_http_header(r);

	if (r->header_only)
		return OK;

	ap_rprintf(
		r, MODULE "/" VERSION
		" Copyright " COPYRIGHT_DATE " by Anthony Howe. All rights reserved.\n"
	);
	ap_rprintf(r, "The official web site including license terms and disclaimer is here:\n");
	ap_rprintf(r, "http://www.snert.com/Software/mod_watch/\n");

	return OK;
}

static void
watchUpdateEntry(struct shEntry *entry, struct shEntry *data)
{
	unsigned long now;

	entry->ifInOctets  += data->ifInOctets;
	entry->ifOutOctets += data->ifOutOctets;
	entry->ifRequests  += data->ifRequests;
	entry->ifDocuments += data->ifDocuments;

	/* Compute an average on the next request after at least N
	 * seconds of real time have passed. Calculation not done
	 * at regular intervals, because frequency of requests is
	 * irregular.
	 */
	now = (unsigned long) time((time_t *) 0);
	entry->periodOctets += data->ifOutOctets;

	if (entry->periodMarker == 0) {
		entry->periodMarker = now;
	} else if (entry->periodMarker + SECONDS_PER_PERIOD <= now) {
		entry->ifOutRate = entry->periodOctets * 1.0 / (now - entry->periodMarker);
		entry->periodMarker = now;
		entry->periodOctets = 0;
	}
}

/***********************************************************************
 ***  Log & Content using Weenie Files
 ***********************************************************************/

static FILE *
watchReadFile(request_rec *r, const char *name, struct shEntry *fill)
{
	int fd;
	FILE *fp;
	const char *file;

	memset(fill, 0, sizeof *fill);
	if ((file = ap_pstrcat(r->pool, spoolDirectoryRun, name, (char *) 0)) == (char *) 0)
		return (FILE *) 0;

	if ((fd = ap_popenf(r->pool, file, O_CREAT|O_RDWR, 0644)) < 0)
		return (FILE *) 0;

	if ((fp = ap_pfdopen(r->pool, fd, "r+")) == (FILE *) 0) {
		ap_pclosef(r->pool, fd);
		return (FILE *) 0;
	}

	fill->key = ap_pstrdup(r->pool, name);

	(void) fscanf(
		fp, shScanFormat,
		&fill->ifInOctets,
		&fill->ifOutOctets,
		&fill->ifRequests,
		&fill->ifDocuments,
		&fill->ifActive,
		&fill->ifOutRate,
		&fill->periodOctets,
		&fill->periodMarker
	);

	(void) chown(file, ap_user_id, ap_group_id);

	return fp;
}

static void
watchPostFiles(request_rec *r, const char *keyNotes, const char *keyHash)
{
	FILE *fp;
	struct shEntry *copy;

	if (keyHash == (char *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, watchMainServer,
			"counters not loaded, bogus hostname for request: %s", r->hostname
		);
		goto error0;
	}

	/* Use of flock() is a problem under SunOS and HP-UX.
	 * Use a real semaphore instead.
	 */
	if (MutexLock(mutex))
		goto error0;

	copy = ap_palloc(r->pool, sizeof *copy);

	if ((fp = watchReadFile(r, keyHash, copy)) == (FILE *) 0)
		goto error1;

	ap_table_setn(r->notes, keyNotes, (char *) copy);

	/*** NOTE that we don't count active connections when using
	 *** storage direct to weenie files, as its too slow and
	 *** cumbersome to read & update the files two or three times
	 *** per request, instead of just once.
	 ***
	 *** Consider restucturing watchLog() as watchCleanUp() to
	 *** get around this.
	 ***/
	copy->ifActive = 1;

	ap_pfclose(r->pool, fp);
error1:
	(void) MutexUnlock(mutex);
error0:
	;
}

static void
watchLogFiles(request_rec *r, const char *file, struct shEntry *data)
{
	FILE *fp;
	struct shEntry entry;

	if (file == (const char *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, watchMainServer,
			"not logged, bogus hostname for request: %s", r->hostname
		);
		return;
	}

	/* Use of flock() is a problem under SunOS and HP-UX.
	 * Use a real semaphore instead.
	 */
	if (MutexLock(mutex))
		goto error0;

	if ((fp = watchReadFile(r, file, &entry)) == (FILE *) 0)
		goto error1;

	watchUpdateEntry(&entry, data);

	rewind(fp);
	(void) fprintf(
		fp, shPrintFormat,
		entry.ifInOctets,
		entry.ifOutOctets,
		entry.ifRequests,
		entry.ifDocuments,
		entry.ifActive,
		entry.ifOutRate,
		entry.periodOctets,
		entry.periodMarker
	);

	(void) fprintf(fp, "\n");

	ap_pfclose(r->pool, fp);
error1:
	(void) MutexUnlock(mutex);
error0:
	;
}

static int
watchContentInfoFiles(request_rec *r)
{
	int rc;
	FILE *fp;
	const char *file;
	struct shEntry fill;

	if ((file = watchTarget(r)) == (const char *) 0)
		return HTTP_NOT_FOUND;

	if ((rc = ap_discard_request_body(r)) != OK)
		return rc;

	if ((fp = watchReadFile(r, file, &fill)) == (FILE *) 0)
		return HTTP_INTERNAL_SERVER_ERROR;

	ap_pfclose(r->pool, fp);

	return watchContentInfo(r, &fill);
}

/***********************************************************************
 ***  Log & Content using Shared Memory Hash
 ***********************************************************************/

static void
watchCleanUpHash(void *data)
{
	char *name = (char *) data;
	struct shEntry *entry = shGetLockedEntry(shtable, name);

	if (entry == (struct shEntry *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_CRIT, watchMainServer,
			"shGetLockedEntry(%lx, \"%s\") failed in watchCleanUpHash()",
			(unsigned long) shtable, name
		);
		return;
	}

	if (--entry->ifActive < 0) {
		ap_log_error(
			APLOG_MARK, APLOG_WARNING, watchMainServer,
			"\"%s\" concurrency counter went negative; resetting to zero",
			name
		);
		entry->ifActive = 0;
	}

	ap_log_error(
		APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, watchMainServer,
		"watchCleanUpHash(%lx) key=%s active=%d",
		data, entry->key, entry->ifActive
	);

	if (shUnlock(shtable)) {
		ap_log_error(
			APLOG_MARK, APLOG_CRIT, watchMainServer,
			"shUnlock(%lx) failed in watchCleanUpHash()",
			(unsigned long) shtable
		);
	}
}

static void
watchPostHash(request_rec *r, const char *keyNotes, const char *keyHash)
{
	struct shEntry *copy, *entry;

	if (keyHash == (char *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, watchMainServer,
			"counters not loaded, bogus hostname for request: %s", r->hostname
		);
		return;
	}

	entry = shGetLockedEntry(shtable, keyHash);
	if (entry == (struct shEntry *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_CRIT, watchMainServer,
			"shGetLockedEntry(%lx, \"%s\") failed in watchPostHash()",
			(unsigned long) shtable, keyHash
		);
		return;
	}

	/* Register a clean-up handler before we count the active
	 * connection. This could lead to a negative value during
	 * cleanup if the connection is dropped between the
	 * ap_register_cleanup() call and the counter increment.
	 *
	 * Reversing the following two lines can result in a very
	 * rare condition where a connection is counted, then drops
	 * between the two lines, and then is never uncounted. At
	 * least the other way we can detect the condition more
	 * readily.
	 *
	 * Of course the connection could be dropped before we even
	 * get this far, which doesn't matter.
	 *
	 * Note that ALL requests, including watch handlers, are
	 * counted in this phase since its too early to detect our
	 * handler. I could do a subrequest to find out, but I see
	 * little benefit in filtering out the handlers.
	 */
	ap_register_cleanup(r->pool, (char *) keyHash, watchCleanUpHash, ap_null_cleanup);
	entry->ifActive++;

	ap_log_error(
		APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, watchMainServer,
		"watchPostHash(%lx, %s, %s) active=%d",
		r, keyNotes, keyHash, entry->ifActive
	);

	copy = ap_palloc(r->pool, sizeof *copy);
	*copy = *entry;
	ap_table_setn(r->notes, keyNotes, (char *) copy);

	if (shUnlock(shtable)) {
		ap_log_error(
			APLOG_MARK, APLOG_CRIT, watchMainServer,
			"shUnlock(%lx) failed in watchPostHash()",
			(unsigned long) shtable
		);
	}
}

static void
watchLogHash(request_rec *r, const char *name, struct shEntry *data)
{
	unsigned long rate;
	struct shEntry *entry;

	if (name == (char *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, watchMainServer,
			"not logged, bogus hostname for request: %s", r->hostname
		);
		return;
	}

	entry = shGetLockedEntry(shtable, name);

	if (entry == (struct shEntry *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_CRIT, watchMainServer,
			"shGetLockedEntry(%lx, \"%s\") failed in watchLogHash()",
			(unsigned long) shtable, name
		);
		return;
	}

	watchUpdateEntry(entry, data);

	if (shUnlock(shtable)) {
		ap_log_error(
			APLOG_MARK, APLOG_CRIT, watchMainServer,
			"shUnlock(%lx) failed in watchLogHash()",
			(unsigned long) shtable
		);
	}
}

static int
watchContentInfoHash(request_rec *r)
{
	int rc;
	const char *name;
	struct shEntry *entry, *copy;

	if ((name = watchTarget(r)) == (const char *) 0)
		return HTTP_NOT_FOUND;

	if (!shContainsKey(shtable, name))
		return HTTP_NOT_FOUND;

	copy = ap_palloc(r->pool, sizeof *copy);

	if ((rc = ap_discard_request_body(r)) != OK)
		return rc;

	entry = shGetLockedEntry(shtable, name);

	if (entry == (struct shEntry *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_CRIT, watchMainServer,
			"shGetLockedEntry(%lx, \"%s\") failed in watchContentInfoHash()",
			(unsigned long) shtable, name
		);

		memset(copy, 0, sizeof *copy);
		copy->key = (char *) name;
	} else {
		*copy = *entry;

		if (shUnlock(shtable)) {
			ap_log_error(
				APLOG_MARK, APLOG_CRIT, watchMainServer,
				"shUnlock(%lx) failed in watchContentInfoHash()",
				(unsigned long) shtable
			);
		}
	}

	return watchContentInfo(r, copy);
}

/***********************************************************************
 ***  List & Table Content.
 ***********************************************************************/

static int
watchContentList(request_rec *r)
{
	int i;
	DIR *dir;
	FILE *fp;
	struct dirent *dirent;
	struct shEntry *array, entry;

	ap_send_http_header(r);

	ap_log_error(
		APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, watchMainServer,
		"watchContentList(%lx) shtable=%lx shtable.mutex=%lx mutex=%lx", r,
		(unsigned long) shtable,
		(unsigned long) (shtable == (struct shTable *) 0 ? 0 : shtable->mutex),
		(unsigned long) mutex
	);

	if (r->header_only)
		goto error0;

	if (shLock(shtable) && MutexLock(mutex))
		goto error0;

	if (shtable != (struct shTable *) 0) {
		array = shtable->array;
		for (i = 0; i < shTableSize; ++i) {
			if (array[i].key == (char *) 0)
				continue;
			watchContentInfoLine(r, &array[i]);
		}
	}

	/* List all weenie files NOT loaded into the hash table. */
	if ((dir = ap_popendir(r->pool, spoolDirectoryRun)) == (DIR *) 0)
		goto error1;

	while ((dirent = readdir(dir)) != (struct dirent *) 0) {
		if (*dirent->d_name == '.' || strcmp(dirent->d_name, shLockFile) == 0)
			continue;

		if (shContainsKey(shtable, dirent->d_name))
			continue;

		if ((fp = watchReadFile(r, dirent->d_name, &entry)) == (FILE *) 0)
			continue;

		ap_pfclose(r->pool, fp);

		watchContentInfoLine(r, &entry);
	}

	ap_pclosedir(r->pool, dir);
error1:
	shUnlock(shtable) && MutexUnlock(mutex);
error0:
	return OK;
}

static int
watchContentTable(request_rec *r)
{
	int i;
	DIR *dir;
	FILE *fp;
	unsigned long hash;
	struct dirent *dirent;
	struct shEntry *array, entry;

	ap_send_http_header(r);

	ap_log_error(
		APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, watchMainServer,
		"watchContentTable(%lx) shtable=%lx shtable.mutex=%lx mutex=%lx", r,
		(unsigned long) shtable,
		(unsigned long) (shtable == (struct shTable *) 0 ? 0 : shtable->mutex),
		(unsigned long) mutex
	);

	if (r->header_only)
		goto error0;

	if (shLock(shtable) && MutexLock(mutex))
		goto error0;

	ap_rprintf(r, "<html>");
	ap_rprintf(r, "<style type='text/css'>");
	ap_rprintf(r, "th {");
	ap_rprintf(r, "  font-family: Verdana, Arial, Helvetica, sans-serif;");
	ap_rprintf(r, "  font-size: 10pt;");
	ap_rprintf(r, "  background-color: #0080D7;");
	ap_rprintf(r, "  color: #ffffff;");
	ap_rprintf(r, "}");
	ap_rprintf(r, ".normal, p, ul, td {");
	ap_rprintf(r, "  font-family: Verdana, Arial, Helvetica, sans-serif;");
	ap_rprintf(r, "  font-size: 10pt;");
	ap_rprintf(r, "  color: #000000;");
	ap_rprintf(r, "}");
	ap_rprintf(r, "</style>");
	ap_rprintf(r, "<body>");

	ap_rprintf(r, "<p align='center'>"MODULE"/"VERSION"</p>\n");

	if (shtable != (struct shTable *) 0) {
		ap_rprintf(r, "<table width='100%%' border='0' cellpadding='0' cellspacing='0'>\n");
		ap_rprintf(r, "<tr align='center'>\n");
		ap_rprintf(r, "<th>Table Size</th><th>Found</th><th>Probes</th><th>Faults</th><th>Flushes</th><th>Shared Memory Remaining</th>\n");
		ap_rprintf(r, "</tr>\n");
		ap_rprintf(r, "<tr align='center'>\n");
		ap_rprintf(
			r, "<td>%d</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%lu</td>\n",
			shtable->size, shtable->info->found, shtable->info->probes,
			shtable->info->faults, shtable->info->flushes,
			MemoryAvailable(shtable->memory)

		);
		ap_rprintf(r, "</tr>\n");
		ap_rprintf(r, "</table>\n");
	}

	ap_rprintf(r, "<table width='100%%' border='0' cellpadding='1' cellspacing='0'>\n");
	ap_rprintf(r, "<tr align='right'>\n");
	ap_rprintf(r, "<th>Key</th><th>Hash</th><th>Optimal</th><th>Index</th><th>InOctets</th><th>OutOctets</th><th>Requests</th><th>Documents</th><th>B/s</th><th>Active</th>\n");
	ap_rprintf(r, "</tr>\n");

	if (shtable != (struct shTable *) 0) {
		array = shtable->array;
		for (i = 0; i < shTableSize; ++i) {
			if (array[i].key == (char *) 0)
				continue;

			hash = shHashCode(0, array[i].key);

			ap_rprintf(r, "<tr align='right'>\n");
			ap_rprintf(
				r, "<td>%s</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%.3f</td><td>%hd</td>\n",
				array[i].key,
				hash,
				hash % shTableSize,
				i,
				array[i].ifInOctets,
				array[i].ifOutOctets,
				array[i].ifRequests,
				array[i].ifDocuments,
				array[i].ifOutRate,
				array[i].ifActive
			);
			ap_rprintf(r, "</tr>\n");
		}
	}

	/* List all weenie files NOT loaded into the hash table. */
	if ((dir = ap_popendir(r->pool, spoolDirectoryRun)) == (DIR *) 0)
		goto error1;

	while ((dirent = readdir(dir)) != (struct dirent *) 0) {
		if (*dirent->d_name == '.' || strcmp(dirent->d_name, shLockFile) == 0)
			continue;

		if (shContainsKey(shtable, dirent->d_name))
			continue;

		if ((fp = watchReadFile(r, dirent->d_name, &entry)) == (FILE *) 0)
			continue;

		ap_pfclose(r->pool, fp);

		hash = shHashCode(0, dirent->d_name);

		ap_rprintf(r, "<tr align='right'>\n");
		ap_rprintf(
			r, "<td>%s</td><td>%lu</td><td>%lu</td><td>--</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%.3f</td><td>%hd</td>\n",
			dirent->d_name,
			hash,
			hash % shTableSize,
			entry.ifInOctets,
			entry.ifOutOctets,
			entry.ifRequests,
			entry.ifDocuments,
			entry.ifOutRate,
			entry.ifActive
		);
		ap_rprintf(r, "</tr>\n");
	}

	ap_pclosedir(r->pool, dir);
error1:
	ap_rprintf(r, "</table>\n");
	ap_rprintf(r, "</body></html>\n");
	shUnlock(shtable) && MutexUnlock(mutex);
error0:
	return OK;
}

static int
watchContentFlush(request_rec *r)
{
	ap_send_http_header(r);

	if (shtable != (struct shTable *) 0) {
		if (!shLock(shtable))
			shFlushAll(shtable);
		shUnlock(shtable);
	}

	ap_rprintf(r, "OK\n");

	return OK;
}

/***********************************************************************
 ***  Configuration Directives
 ***********************************************************************/

/*
 * Allows the URL used to query virtual host data:
 *
 *	http://www.snert.com/watch-info
 *
 *	<Location /watch-info>
 *	SetHandler watch-info
 *	</Location>
 *
 *	http://www.snert.com/watch-list
 *
 *	<Location /watch-list>
 *	SetHandler watch-list
 *	</Location>
 *
 * Allows the URL used to query file owner and web
 * server data:
 *
 *	http://www.snert.com/~achowe/watch-info
 *	http://www.snert.com/~SERVER/watch-info
 *
 *	<Location /~* /watch-info>
 *	SetHandler watch-info
 *	</Location>
 *
 *	<Location /~* /watch-info>
 *	SetHandler watch-info
 *	</Location>
 *
 * The following allows me to verify the version and copyright:
 *
 *	<Location /watch-copyright>
 *	SetHandler watch-copyright
 *	Order allow,deny
 *	Allow from snert.biz snert.com snert.info snert.net snert.org
 *	</Location>
 */
static handler_rec watchHandlers[] = {
	{ watch_info, watchContentInfoHash },
	{ watch_list, watchContentList },
	{ watch_table, watchContentTable },
	{ watch_flush, watchContentFlush },
	{ watch_copyright, watchContentCopyright },
	{ NULL }
};

/*
 * WatchHashTableSize number
 *
 * The size of the shared memory hash table for the "hash" storage policy.
 */
static const char *
WatchHashTableSize(cmd_parms *cmd, void *dconf, char *number)
{
	return setUnsignedInt(number, &shTableSize);
}

/*
 * WatchSpoolDirectory directory
 *
 * Absolute or server root relative directory where support and runtime
 * files are kept.
 */
static const char *
WatchSpoolDirectory(cmd_parms *cmd, void *dconf, char *directory)
{
	return setDirectory(cmd->pool, directory, &spoolDirectoryConf);
}

/*
 * WatchStorage files|hash
 */
static const char *
WatchStorage(cmd_parms *cmd, void *dconf, const char *policy)
{
	long value;
	storage_policy *dp;
	const char *msg;

	for (dp = policy_table; dp->name != (const char *) 0; ++dp) {
		if (ap_strcasecmp_match(policy, dp->name) == 0)
			break;
	}

	if (dp->name == (const char *) 0)
		return "Invalid storage policy.";

	/* Set the handler to be used for logging. */
	watchStoragePolicy = dp;
	watchHandlers[0].handler = dp->content;

	return (const char *) 0;
}

/*
 * WatchDocuments mime-type ...
 */
static const char *
WatchDocuments(cmd_parms *cmd, void *dconf, char *arg)
{
	char **mime;

	if (watchDocumentsConf == (array_header *) 0)
		watchDocumentsConf = ap_make_array(cmd->pool, 0, sizeof *watchDocumentsDef);

	mime = (char **) ap_push_array(watchDocumentsConf);
	if (mime == (char **) 0)
		return "Cannot add to WatchDocuments list.";

	*mime = (char *) ap_pstrdup(cmd->pool, arg);

	return (const char *) 0;
}

/*
 * WatchSpoolFile unique-filename
 */
static const char *
WatchSpoolFile(cmd_parms *cmd, void *dconf, char *filename)
{
	int i;
	char **entry, **table;

	if (watchSpoolFileConf == (array_header *) 0)
		watchSpoolFileConf = ap_make_array(cmd->pool, 0, sizeof (char *));

	table = (char **) watchSpoolFileConf->elts;
	for (i = 0; i < watchSpoolFileConf->nelts; ++i) {
		if (strcmp(filename, table[i]) == 0)
			return "WatchSpoolFile argument is not unique.";
	}

	((struct watchConfDir *) dconf)->index = watchSpoolFileConf->nelts;

	entry = (char **) ap_push_array(watchSpoolFileConf);
	if (entry == (char **) 0)
		return "Cannot add to WatchSpoolFile list.";

	*entry = (char *) ap_pstrdup(cmd->pool, filename);

	return (const char *) 0;
}

/*
 * WatchWhat item ...
 */
static const char *
WatchWhat(cmd_parms *cmd, void *dconf, char *arg)
{
	/* Erase the default settings when explicitly set. */
	if (watch_what & WATCH_DEFAULT)
		watch_what = 0;

	if (ap_strcasecmp_match(arg, "nothing") == 0)
		watch_what = 0;
	else if (ap_strcasecmp_match(arg, "virtual-host") == 0)
		watch_what |= WATCH_VHOST;
	else if (ap_strcasecmp_match(arg, "virtual-host-by-port") == 0)
		watch_what |= WATCH_VHOST_BY_PORT;
	else if (ap_strcasecmp_match(arg, "file-owner") == 0)
		watch_what |= WATCH_OWNER;
	else if (ap_strcasecmp_match(arg, "file-owner-by-port") == 0)
		watch_what |= WATCH_OWNER_BY_PORT;
	else if (ap_strcasecmp_match(arg, "server") == 0)
		watch_what |= WATCH_SERVER;
	else if (ap_strcasecmp_match(arg, "server-by-port") == 0)
		watch_what |= WATCH_SERVER_BY_PORT;
	else if (ap_strcasecmp_match(arg, "remote-ip") == 0)
		watch_what |= WATCH_REMOTE_IP;
	else if (ap_strcasecmp_match(arg, "remote-ip-by-port") == 0)
		watch_what |= WATCH_REMOTE_IP_BY_PORT;
	else if (ap_strcasecmp_match(arg, "headers-in") == 0)
		watch_what |= WATCH_HEADERS_IN;
	else if (ap_strcasecmp_match(arg, "headers-out") == 0)
		watch_what |= WATCH_HEADERS_OUT;
	else if (ap_strcasecmp_match(arg, "content-length") == 0)
		watch_what |= WATCH_CONTENT_LENGTH;
	else if (ap_strcasecmp_match(arg, watch_all) == 0)
		watch_what |= (WATCH_ALL & ~WATCH_DEFAULT);
	else
		return "Invalid item for WatchWhat";

	return (const char *) 0;
}

/*
 * WatchNotLogged handler ...
 */
static const char *
WatchNotLogged(cmd_parms *cmd, void *dconf, char *arg)
{
	/* Erase the default settings when explicitly set. */
	if (watch_not_logged == ~0)
		watch_not_logged = 0;

	if (ap_strcasecmp_match(arg, watch_copyright) == 0)
		watch_not_logged |= WATCH_COPYRIGHT;
	else if (ap_strcasecmp_match(arg, watch_flush) == 0)
		watch_not_logged |= WATCH_FLUSH;
	else if (ap_strcasecmp_match(arg, watch_info) == 0)
		watch_not_logged |= WATCH_INFO;
	else if (ap_strcasecmp_match(arg, watch_list) == 0)
		watch_not_logged |= WATCH_LIST;
	else if (ap_strcasecmp_match(arg, watch_table) == 0)
		watch_not_logged |= WATCH_TABLE;
	else if (ap_strcasecmp_match(arg, watch_flush) == 0)
		watch_not_logged |= WATCH_FLUSH;
	else if (ap_strcasecmp_match(arg, watch_all) == 0)
		watch_not_logged |= WATCH_COPYRIGHT|WATCH_INFO|WATCH_LIST|WATCH_TABLE|WATCH_FLUSH;
	else
		return "Invalid handler name for WatchNotLogged";

	return (const char *) 0;
}

/*
 * WatchNetwork {include|exclude} ip/mask ...
 */
static const char *
WatchNetwork(cmd_parms *cmd, void *dconf, char *set, char *network)
{
	if (networkIncludeConf == (array_header *) 0)
		networkIncludeConf = ap_make_array(cmd->pool, 0, sizeof (struct network));

	if (networkExcludeConf == (array_header *) 0)
		networkExcludeConf = ap_make_array(cmd->pool, 0, sizeof (struct network));

	if (ap_strcasecmp_match(set, "include") == 0)
		return ntAddNetwork(networkIncludeConf, network);
	else if (ap_strcasecmp_match(set, "exclude") == 0)
		return ntAddNetwork(networkExcludeConf, network);

	return "Invalid category for WatchNetwork";
}

static command_rec watchCommands[] = {
	{ "WatchDocuments", WatchDocuments, NULL, RSRC_CONF, ITERATE,
	"List of MIME types considered to be documents." },

	{ "WatchSpoolFile", WatchSpoolFile, NULL, ACCESS_CONF, TAKE1,
	"A unique filename per <Directory> or <Location> that is to be watched." },

	{ "WatchHashTableSize", WatchHashTableSize, NULL, RSRC_CONF, TAKE1,
	"Shared memory hash table size for \"hash\" storage policy." },

	{ "WatchNotLogged", WatchNotLogged, NULL, RSRC_CONF, ITERATE,
	"Do not log one or more handlers: all, watch-info, watch-list, watch-copyright, watch-table" },

	{ "WatchNetwork", WatchNetwork, NULL, RSRC_CONF, ITERATE2,
	"Include or exclude networks (IP/mask) from being watched." },

	{ "WatchSpoolDirectory", WatchSpoolDirectory, NULL, RSRC_CONF, TAKE1,
	"Spool directory for any support and runtime files." },

	{ "WatchStorage", WatchStorage, NULL, RSRC_CONF, TAKE1,
	"The storage policy: files, hash" },

	{ "WatchWhat", WatchWhat, NULL, RSRC_CONF, ITERATE,
	"One or more items to watch: all, virtual-host, file-owner, remote-ip, server, ..." },

	{ NULL }
};

/***********************************************************************
 ***  Module Definition Block
 ***********************************************************************/

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT watch_module = {
	STANDARD_MODULE_STUFF,
	watchInit,		/* module initializer                  */
	watchCreateDir,		/* create per-dir    conf structures */
	NULL,			/* merge  per-dir    conf structures */
	NULL,			/* create per-server conf structures */
	NULL,			/* merge  per-server conf structures */
	watchCommands,		/* table of conf file commands       */
	watchHandlers,		/* [#8] MIME-typed-dispatched handlers */
	NULL,			/* [#1] URI to filename translation    */
	NULL,			/* [#4] validate user id from request  */
	NULL,			/* [#5] check if the user is ok _here_ */
	NULL,			/* [#3] check access by host address   */
	NULL,			/* [#6] determine MIME type            */
	watchFixup,		/* [#7] pre-run fixups                 */
	watchLog,		/* [#9] log a transaction              */
#if MODULE_MAGIC_NUMBER >= 19970103
	NULL,			/* [#2] header parser                  */
#endif
#if MODULE_MAGIC_NUMBER >= 19970719
	NULL,			/* child_init			       */
#endif
#if MODULE_MAGIC_NUMBER >= 19970728
	NULL,			/* child_exit			       */
#endif
#if MODULE_MAGIC_NUMBER >= 19970902
	watchPost		/* [#0] post read-request              */
#endif
};

